==> v1.10.4.zip: wcs/wcs/qommon/liberty.py <==
# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import urllib
import urlparse
import httplib
import sys

from quixote import get_field, get_request, get_response, get_session, \
        get_session_manager, redirect, get_publisher
from quixote.directory import Directory
from quixote.http_request import parse_header

from publisher import get_cfg, get_logger

try:
    import lasso
except ImportError:
    lasso = None

import misc
from form import *
from template import error_page
import errors

def does_idp_authentication():
    methods = get_cfg('identification', {}).get('methods')
    return 'idp' in methods

class LibertyDirectory(Directory):
    _q_exports = ['login', 'assertionConsumer', 'soapEndpoint',
            'singleLogout', 'singleLogoutReturn',
            'federationTermination', 'federationTerminationReturn',
            ('metadata.xml', 'metadata'), 'public_key']

    def _q_traverse(self, path):
        # if lasso is not installed, hide the liberty endpoints
        if lasso is None:
            if does_idp_authentication():
                rel_path = os.path.join('/liberty', *path)
                get_logger().error(
                        '%s unavailable - lasso is not installed' % rel_path)
            raise errors.TraversalError()
        return Directory._q_traverse(self, path)

    def perform_login(self, idp = None):
        server = misc.get_lasso_server()
        login = lasso.Login(server)
        login.initAuthnRequest(idp, lasso.HTTP_METHOD_REDIRECT)
        login.request.nameIdPolicy = 'federated'
        login.request.forceAuthn = False
        login.request.isPassive = False
        login.request.consent = 'urn:liberty:consent:obtained'
        login.buildAuthnRequestMsg()
        return redirect(login.msgUrl)

    def assertionConsumer(self):
        server = misc.get_lasso_server()
        if not server:
            return error_page(_('Liberty support is not yet configured'))
        login = lasso.Login(server)
        request = get_request()
        if request.get_method() == 'GET' or get_field('LAREQ'):
            if request.get_method() == 'GET':
                try:
                    login.initRequest(request.get_query(), lasso.HTTP_METHOD_REDIRECT)
                except lasso.Error, error:
                    return error_page(_('Failed to parse query'))
            else:
                login.initRequest(get_field('LAREQ'), lasso.HTTP_METHOD_POST)

            login.buildRequestMsg()
            remote_provider_cfg = get_cfg('idp', {}).get(misc.get_provider_key(login.remoteProviderId))
            client_cert = remote_provider_cfg.get('clientcertificate')
            try:
                soap_answer = soap_call(login.msgUrl, login.msgBody, client_cert = client_cert)
            except SOAPException:
                return error_page(_('Failure to communicate with identity provider'))
            try:
                login.processResponseMsg(soap_answer)
            except lasso.Error, error:
                if error[0] == lasso.LOGIN_ERROR_STATUS_NOT_SUCCESS:
                    return error_page(_('Unknown authentication failure'))
                if hasattr(lasso, 'LOGIN_ERROR_UNKNOWN_PRINCIPAL'):
                    if error[0] == lasso.LOGIN_ERROR_UNKNOWN_PRINCIPAL:
                        return error_page(_('Authentication failure; unknown principal'))
                if error[0] == lasso.LOGIN_ERROR_FEDERATION_NOT_FOUND:
                    return error_page('there was no federation')
                return error_page(_("Identity Provider didn't accept artifact transaction."))
        else:
            login.processAuthnResponseMsg(get_field('LARES'))
        login.acceptSso()
        session = get_session()
        if login.isSessionDirty:
            if login.session:
                session.lasso_session_dump = login.session.dump()
            else:
                session.lasso_session_dump = None

        user = self.lookup_user(session, login)
        if user:
            session.set_user(user.id)
        else:
            session.set_user('anonymous-%s' % login.nameIdentifier.content)
            session.lasso_anonymous_identity_dump = login.identity.dump()
        session.lasso_identity_provider_id = login.remoteProviderId

        response = get_response()
        if session.after_url:
            after_url = session.after_url
            session.after_url = None
            return redirect(after_url)
        response.set_status(303)
        response.headers['location'] = urlparse.urljoin(request.get_url(), str('..'))
        response.content_type = 'text/plain'
        return 'Your browser should redirect you'

    def lookup_user(self, session, login):
        ni = login.nameIdentifier.content
        session.name_identifier = ni
        nis = list(get_publisher().user_class.select(lambda x: ni in x.name_identifiers))
        if nis:
            user = nis[0]
        else:
            return None
        user.lasso_dump = login.identity.dump()
        user.store()
        return user

    def singleLogout(self):
        request = get_request()
        logout = lasso.Logout(misc.get_lasso_server())
        if lasso.isLibertyQuery(request.get_query()):
            request = get_request()
            try:
                logout.processRequestMsg(request.get_query())
            except lasso.Error, error:
                if error[0] == lasso.DS_ERROR_INVALID_SIGNATURE:
                    return error_page(_('Failed to check single logout request signature.'))
                raise
            session = get_session()
            if not session.id:
                # session has not been found, this may be because the user has
                # its browser configured so that cookies are not sent for
                # remote queries and IdP is using image-based SLO.
                # so we look up a session with the appropriate name identifier
                name_identifier = logout.nameIdentifier.content
                sessions = get_session_manager().get_session_for_liberty(
                        name_identifier, logout.request.sessionIndex)
                if sessions:
                    session = session[0]
                else:
                    session = get_session()
            return self.slo_idp(logout, session)
        else:
            return self.slo_sp(logout, get_session())

    def singleLogoutReturn(self):
        logout = lasso.Logout(misc.get_lasso_server())
        try:
            logout.processResponseMsg(get_request().get_query())
        except lasso.Error, error:
            if error[0] == lasso.PROFILE_ERROR_INVALID_QUERY:
                raise AccessError()
            if error[0] == lasso.DS_ERROR_INVALID_SIGNATURE:
                return error_page(_('Failed to check single logout request signature.'))
            if hasattr(lasso, 'LOGOUT_ERROR_REQUEST_DENIED') and \
                    error[0] == lasso.LOGOUT_ERROR_REQUEST_DENIED:
                # ignore silently
                return redirect(get_publisher().get_root_url())
            elif error[0] == lasso.ERROR_UNDEFINED:
                # XXX: unknown status; ignoring for now.
                return redirect(get_publisher().get_root_url())
        return redirect(get_publisher().get_root_url())

    def slo_idp(self, logout, session):
        # Single Logout initiated by IdP
        if session and session.lasso_session_dump:
            logout.setSessionFromDump(session.lasso_session_dump)
        user = None
        if session:
            user = session.get_user()
        if user and user.lasso_dump:
            logout.setIdentityFromDump(user.lasso_dump)
        if session and logout.nameIdentifier.content != session.name_identifier:
            raise "no appropriate name identifier in session (%s and %s)" % (
                    logout.nameIdentifier.content, session.name_identifier)

        try:
            logout.validateRequest()
        except lasso.Error, error:
            if error[0] == lasso.PROFILE_ERROR_SESSION_NOT_FOUND:
                pass
            elif error[0] == lasso.PROFILE_ERROR_IDENTITY_NOT_FOUND:
                pass
            else:
                raise
        else:
            get_session_manager().expire_session()
            # Expire all linked sessions
            sessions = get_session_manager().get_session_for_liberty(
                    logout.nameIdentifier.content, logout.request.sessionIndex)
            for session in sessions:
                try:
                    del get_session_manager()[session.id]
                except KeyError:
                    pass

        logout.buildResponseMsg()
        if logout.msgBody: # soap answer
            return logout.msgBody
        else:
            return redirect(logout.msgUrl)

    def slo_sp(self, logout, session):
        if not session.user:
            get_session_manager().expire_session()
            return redirect(get_publisher().get_root_url())

        if session.lasso_session_dump:
            logout.setSessionFromDump(session.lasso_session_dump)
        user = get_request().user
        if user and user.lasso_dump:
            logout.setIdentityFromDump(user.lasso_dump)
        return self.slo_sp_redirect(logout)

    def slo_sp_redirect(self, logout):
        try:
            logout.initRequest(None, lasso.HTTP_METHOD_REDIRECT)
        except lasso.Error, error:
            if error[0] == lasso.PROFILE_ERROR_NAME_IDENTIFIER_NOT_FOUND:
                get_session_manager().expire_session()
                return redirect(get_publisher().get_root_url())
            if error[0] == lasso.PROFILE_ERROR_SESSION_NOT_FOUND:
                get_session_manager().expire_session()
                return redirect(get_publisher().get_root_url())
            raise
        logout.buildRequestMsg()
        get_session_manager().expire_session()
        return redirect(logout.msgUrl)

    def soapEndpoint(self):
        request = get_request()
        ctype = request.environ.get('CONTENT_TYPE')
        if not ctype:
            return # XXX: error code

        ctype, ctype_params = parse_header(ctype)
        if ctype != 'text/xml':
            return # XXX: error code

        response = get_response()
        response.set_content_type('text/xml')

        length = int(request.environ.get('CONTENT_LENGTH'))
        soap_message = request.stdin.read(length)

        request_type = lasso.getRequestTypeFromSoapMsg(soap_message)

        if request_type == lasso.REQUEST_TYPE_LOGOUT:
            logout = lasso.Logout(misc.get_lasso_server())
            logout.processRequestMsg(soap_message)
            name_identifier = logout.nameIdentifier.content
            sessions = get_session_manager().get_session_for_liberty(
                    name_identifier, logout.request.sessionIndex)
            if sessions:
                session = sessions[0]
            else:
                session = None
            return self.slo_idp(logout, session)

        if request_type == lasso.REQUEST_TYPE_DEFEDERATION:
            defederation = lasso.Defederation(misc.get_lasso_server())
            defederation.processNotificationMsg(soap_message)
            name_identifier = defederation.nameIdentifier.content
            user = None
            for session in get_session_manager().values():
                if name_identifier == session.name_identifier:
                    break
            else:
                nis = list(get_publisher().user_class.select(
                            lambda x: name_identifier in x.name_identifiers))
                if not nis:
                    raise 'federation not found'
                user = nis[0]
                session = None
            return self.fedterm(defederation, session, user)

        response.set_status(501)
        request.soap_request_type = request_type
        request.soap_message = soap_message
        return 'Unimplemented SOAP method'


    def federationTermination(self):
        request = get_request()
        if not lasso.isLibertyQuery(request.get_query()):
            return redirect('.')

        defederation = lasso.Defederation(misc.get_lasso_server())
        defederation.processNotificationMsg(request.get_query())
        session = get_session()
        return self.fedterm(defederation, session)

    def fedterm(self, defederation, session, user = None):
        if session:
            defederation.setSessionFromDump(session.lasso_session_dump)

        if not user:
            user = session.get_user()

        if user and user.lasso_dump:
            defederation.setIdentityFromDump(user.lasso_dump)

        try:
            defederation.validateNotification()
        except lasso.Error, error:
            pass # ignore failure (?)
        else:
            if not defederation.identity:
                # if it was the last federation the whole identity dump collapsed
                user.lasso_dump = None
            else:
                user.lasso_dump = defederation.identity.dump()
            user.store()

        if defederation.isSessionDirty:
            if defederation.session:
                session.lasso_session_dump = defederation.session.dump()
            else:
                session.lasso_session_dump = None

        if defederation.msgUrl:
            return redirect(defederation.msgUrl)
        else:
            get_session_manager().commit_changes(session)
            response = get_response()
            response.set_status(204)
            return ''

    def federationTerminationReturn(self):
        return redirect(get_publisher().get_root_url())

    def metadata(self):
        try:
            metadata = unicode(open(misc.get_abs_path(
                            get_cfg('sp')['metadata'])).read(), 'utf-8')
        except KeyError:
            raise errors.TraversalError()
        response = get_response()
        response.set_content_type('text/xml', 'utf-8')
        return metadata.encode('utf-8')

    def public_key(self):
        response = get_response()
        response.set_content_type('application/octet-stream')
        publickey = open(misc.get_abs_path(get_cfg('sp')['publickey'])).read()
        return publickey



class SOAPException(Exception):
    url = None
    def __init__(self, url):
        self.url = url

def soap_call(url, msg, client_cert = None):
    if url.startswith('http://'):
        host, query = urllib.splithost(url[5:])
        conn = httplib.HTTPConnection(host)
    else:
        host, query = urllib.splithost(url[6:])
        conn = httplib.HTTPSConnection(host,
                key_file = client_cert, cert_file = client_cert)
    try:
        conn.request('POST', query, msg, {'Content-Type': 'text/xml'})
        response = conn.getresponse()
    except Exception, err:
        # exception could be raised by request
        get_logger().warn('SOAP error (on %s): %s' % (url, err))
        raise SOAPException(url)
    data = response.read()
    conn.close()
    if response.status not in (200, 204): # 204 ok for federation termination
        get_logger().warn('SOAP error (%s) (on %s)' % (response.status, url))
        raise SOAPException(url)
    return data

==> v1.12.7.zip: wcs/wcs/qommon/liberty.py <==
# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import urllib
import urlparse
import httplib

from quixote import get_field, get_request, get_response, get_session, \
        get_session_manager, redirect, get_publisher
from quixote.directory import Directory
from quixote.http_request import parse_header

from publisher import get_cfg, get_logger

try:
    import lasso
except ImportError:
    lasso = None

import misc
from form import *
from template import error_page
import errors

def does_idp_authentication():
    methods = get_cfg('identification', {}).get('methods')
    return 'idp' in methods

class LibertyDirectory(Directory):
    _q_exports = ['login', 'assertionConsumer', 'soapEndpoint',
            'singleLogout', 'singleLogoutReturn',
            'federationTermination', 'federationTerminationReturn',
            ('metadata.xml', 'metadata'), 'public_key']

    def _q_traverse(self, path):
        # if lasso is not installed, hide the liberty endpoints
        if lasso is None:
            if does_idp_authentication():
                rel_path = os.path.join('/liberty', *path)
                get_logger().error(
                        '%s unavailable - lasso is not installed' % rel_path)
            raise errors.TraversalError()
        return Directory._q_traverse(self, path)

    def perform_login(self, idp = None):
        server = misc.get_lasso_server()
        login = lasso.Login(server)
        login.initAuthnRequest(idp, lasso.HTTP_METHOD_REDIRECT)
        login.request.nameIdPolicy = 'federated'
        login.request.forceAuthn = False
        login.request.isPassive = False
        login.request.consent = 'urn:liberty:consent:obtained'
        login.buildAuthnRequestMsg()
        return redirect(login.msgUrl)

    def assertionConsumer(self):
        server = misc.get_lasso_server()
        if not server:
            return error_page(_('Liberty support is not yet configured'))
        login = lasso.Login(server)
        request = get_request()
        if request.get_method() == 'GET' or get_field('LAREQ'):
            if request.get_method() == 'GET':
                try:
                    login.initRequest(request.get_query(), lasso.HTTP_METHOD_REDIRECT)
                except lasso.Error, error:
                    return error_page(_('Failed to parse query'))
            else:
                login.initRequest(get_field('LAREQ'), lasso.HTTP_METHOD_POST)

            login.buildRequestMsg()
            remote_provider_cfg = get_cfg('idp', {}).get(misc.get_provider_key(login.remoteProviderId))
            client_cert = remote_provider_cfg.get('clientcertificate')
            try:
                soap_answer = soap_call(login.msgUrl, login.msgBody, client_cert = client_cert)
            except SOAPException:
                return error_page(_('Failure to communicate with identity provider'))
            try:
                login.processResponseMsg(soap_answer)
            except lasso.Error, error:
                if error[0] == lasso.LOGIN_ERROR_STATUS_NOT_SUCCESS:
                    return error_page(_('Unknown authentication failure'))
                if hasattr(lasso, 'LOGIN_ERROR_UNKNOWN_PRINCIPAL'):
                    if error[0] == lasso.LOGIN_ERROR_UNKNOWN_PRINCIPAL:
                        return error_page(_('Authentication failure; unknown principal'))
                if error[0] == lasso.LOGIN_ERROR_FEDERATION_NOT_FOUND:
                    return error_page('there was no federation')
                return error_page(_("Identity Provider didn't accept artifact transaction."))
        else:
            login.processAuthnResponseMsg(get_field('LARES'))
        login.acceptSso()
        session = get_session()
        if login.isSessionDirty:
            if login.session:
                session.lasso_session_dump = login.session.dump()
            else:
                session.lasso_session_dump = None

        user = self.lookup_user(session, login)
        if user:
            session.set_user(user.id)
        else:
            session.set_user('anonymous-%s' % login.nameIdentifier.content)
            session.lasso_anonymous_identity_dump = login.identity.dump()
        session.lasso_identity_provider_id = login.remoteProviderId

        response = get_response()
        if session.after_url:
            after_url = session.after_url
            session.after_url = None
            return redirect(after_url)
        response.set_status(303)
        response.headers['location'] = urlparse.urljoin(request.get_url(), str('..'))
        response.content_type = 'text/plain'
        return 'Your browser should redirect you'

    def lookup_user(self, session, login):
        ni = login.nameIdentifier.content
        session.name_identifier = ni
        nis = list(get_publisher().user_class.get_users_with_name_identifier(ni))
        if nis:
            user = nis[0]
        else:
            return None
        user.lasso_dump = login.identity.dump()
        user.store()
        return user

    def singleLogout(self):
        request = get_request()
        logout = lasso.Logout(misc.get_lasso_server())
        if lasso.isLibertyQuery(request.get_query()):
            request = get_request()
            try:
                logout.processRequestMsg(request.get_query())
            except lasso.Error, error:
                if error[0] == lasso.DS_ERROR_INVALID_SIGNATURE:
                    return error_page(_('Failed to check single logout request signature.'))
                raise
            session = get_session()
            if not session.id:
                # session has not been found, this may be because the user has
                # its browser configured so that cookies are not sent for
                # remote queries and IdP is using image-based SLO.
                # so we look up a session with the appropriate name identifier
                name_identifier = logout.nameIdentifier.content
                sessions = get_session_manager().get_session_for_liberty(
                        name_identifier, logout.request.sessionIndex)
                if sessions:
                    session = session[0]
                else:
                    session = get_session()
            return self.slo_idp(logout, session)
        else:
            return self.slo_sp(logout, get_session())

    def singleLogoutReturn(self):
        logout = lasso.Logout(misc.get_lasso_server())
        try:
            logout.processResponseMsg(get_request().get_query())
        except lasso.Error, error:
            if error[0] == lasso.PROFILE_ERROR_INVALID_QUERY:
                raise AccessError()
            if error[0] == lasso.DS_ERROR_INVALID_SIGNATURE:
                return error_page(_('Failed to check single logout request signature.'))
            if hasattr(lasso, 'LOGOUT_ERROR_REQUEST_DENIED') and \
                    error[0] == lasso.LOGOUT_ERROR_REQUEST_DENIED:
                # ignore silently
                return redirect(get_publisher().get_root_url())
            elif error[0] == lasso.ERROR_UNDEFINED:
                # XXX: unknown status; ignoring for now.
                return redirect(get_publisher().get_root_url())
        return redirect(get_publisher().get_root_url())

    def slo_idp(self, logout, session):
        # Single Logout initiated by IdP
        if session and session.lasso_session_dump:
            logout.setSessionFromDump(session.lasso_session_dump)
        user = None
        if session:
            user = session.get_user()
        if user and user.lasso_dump:
            logout.setIdentityFromDump(user.lasso_dump)
        if session and logout.nameIdentifier.content != session.name_identifier:
            raise Exception("no appropriate name identifier in session (%s and %s)" % (
                    logout.nameIdentifier.content, session.name_identifier))

        try:
            logout.validateRequest()
        except lasso.Error, error:
            if error[0] == lasso.PROFILE_ERROR_SESSION_NOT_FOUND:
                pass
            elif error[0] == lasso.PROFILE_ERROR_IDENTITY_NOT_FOUND:
                pass
            else:
                raise
        else:
            get_session_manager().expire_session()
            # Expire all linked sessions
            sessions = get_session_manager().get_session_for_liberty(
                    logout.nameIdentifier.content, logout.request.sessionIndex)
            for session in sessions:
                try:
                    del get_session_manager()[session.id]
                except KeyError:
                    pass

        logout.buildResponseMsg()
        if logout.msgBody: # soap answer
            return logout.msgBody
        else:
            return redirect(logout.msgUrl)

    def slo_sp(self, logout, session):
        if not session.user:
            get_session_manager().expire_session()
            return redirect(get_publisher().get_root_url())

        if session.lasso_session_dump:
            logout.setSessionFromDump(session.lasso_session_dump)
        user = get_request().user
        if user and user.lasso_dump:
            logout.setIdentityFromDump(user.lasso_dump)
        return self.slo_sp_redirect(logout)

    def slo_sp_redirect(self, logout):
        try:
            logout.initRequest(None, lasso.HTTP_METHOD_REDIRECT)
        except lasso.Error, error:
            if error[0] == lasso.PROFILE_ERROR_NAME_IDENTIFIER_NOT_FOUND:
                get_session_manager().expire_session()
                return redirect(get_publisher().get_root_url())
            if error[0] == lasso.PROFILE_ERROR_SESSION_NOT_FOUND:
                get_session_manager().expire_session()
                return redirect(get_publisher().get_root_url())
            raise
        logout.buildRequestMsg()
        get_session_manager().expire_session()
        return redirect(logout.msgUrl)

    def soapEndpoint(self):
        request = get_request()
        ctype = request.environ.get('CONTENT_TYPE')
        if not ctype:
            return # XXX: error code

        ctype, ctype_params = parse_header(ctype)
        if ctype != 'text/xml':
            return # XXX: error code

        response = get_response()
        response.set_content_type('text/xml')

        length = int(request.environ.get('CONTENT_LENGTH'))
        soap_message = request.stdin.read(length)

        request_type = lasso.getRequestTypeFromSoapMsg(soap_message)

        if request_type == lasso.REQUEST_TYPE_LOGOUT:
            logout = lasso.Logout(misc.get_lasso_server())
            logout.processRequestMsg(soap_message)
            name_identifier = logout.nameIdentifier.content
            sessions = get_session_manager().get_session_for_liberty(
                    name_identifier, logout.request.sessionIndex)
            if sessions:
                session = sessions[0]
            else:
                session = None
            return self.slo_idp(logout, session)

        if request_type == lasso.REQUEST_TYPE_DEFEDERATION:
            defederation = lasso.Defederation(misc.get_lasso_server())
            defederation.processNotificationMsg(soap_message)
            name_identifier = defederation.nameIdentifier.content
            user = None
            for session in get_session_manager().values():
                if name_identifier == session.name_identifier:
                    break
            else:
                nis = list(get_publisher().user_class.get_users_with_name_identifier(
                    name_identifier))
                if not nis:
                    raise Exception('federation not found')
                user = nis[0]
                session = None
            return self.fedterm(defederation, session, user)

        response.set_status(501)
        request.soap_request_type = request_type
        request.soap_message = soap_message
        return 'Unimplemented SOAP method'


    def federationTermination(self):
        request = get_request()
        if not lasso.isLibertyQuery(request.get_query()):
            return redirect('.')

        defederation = lasso.Defederation(misc.get_lasso_server())
        defederation.processNotificationMsg(request.get_query())
        session = get_session()
        return self.fedterm(defederation, session)

    def fedterm(self, defederation, session, user = None):
        if session:
            defederation.setSessionFromDump(session.lasso_session_dump)

        if not user:
            user = session.get_user()

        if user and user.lasso_dump:
            defederation.setIdentityFromDump(user.lasso_dump)

        try:
            defederation.validateNotification()
        except lasso.Error, error:
            pass # ignore failure (?)
        else:
            if not defederation.identity:
                # if it was the last federation the whole identity dump collapsed
                user.lasso_dump = None
            else:
                user.lasso_dump = defederation.identity.dump()
            user.store()

        if defederation.isSessionDirty:
            if defederation.session:
                session.lasso_session_dump = defederation.session.dump()
            else:
                session.lasso_session_dump = None

        if defederation.msgUrl:
            return redirect(defederation.msgUrl)
        else:
            get_session_manager().commit_changes(session)
            response = get_response()
            response.set_status(204)
            return ''

    def federationTerminationReturn(self):
        return redirect(get_publisher().get_root_url())

    def metadata(self):
        try:
            metadata = unicode(open(misc.get_abs_path(
                            get_cfg('sp')['metadata'])).read(), 'utf-8')
        except KeyError:
            raise errors.TraversalError()
        response = get_response()
        response.set_content_type('text/xml', 'utf-8')
        return metadata.encode('utf-8')

    def public_key(self):
        response = get_response()
        response.set_content_type('application/octet-stream')
        publickey = open(misc.get_abs_path(get_cfg('sp')['publickey'])).read()
        return publickey



class SOAPException(Exception):
    url = None
    def __init__(self, url):
        self.url = url

def soap_call(url, msg, client_cert = None):
    if url.startswith('http://'):
        host, query = urllib.splithost(url[5:])
        conn = httplib.HTTPConnection(host)
    else:
        host, query = urllib.splithost(url[6:])
        conn = httplib.HTTPSConnection(host,
                key_file = client_cert, cert_file = client_cert)
    try:
        conn.request('POST', query, msg, {'Content-Type': 'text/xml'})
        response = conn.getresponse()
    except Exception, err:
        # exception could be raised by request
        get_logger().warn('SOAP error (on %s): %s' % (url, err))
        raise SOAPException(url)
    data = response.read()
    conn.close()
    if response.status not in (200, 204): # 204 ok for federation termination
        get_logger().warn('SOAP error (%s) (on %s)' % (response.status, url))
        raise SOAPException(url)
    return data
